SystemOrganization addCategory: #'Neo-Unicode-Sorting'!

Trait named: #TCollateAccess
	uses: {}
	category: 'Neo-Unicode-Sorting'!

!TCollateAccess methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:55'!
collateSequenceAt: anIndex in: anNKDString
	| sequence nextSequence nextIndex |
	sequence := self
		collateSequenceFor: (anNKDString at: anIndex ifAbsent: nil).
	nextIndex := anIndex + 1.
	[ nextIndex <= anNKDString size
		and:
			[ nextSequence := sequence first
				combinedWith: (anNKDString at: nextIndex) codePoint ] ]
		whileTrue: [ sequence := nextSequence.
			nextIndex := anIndex + 1 ].
	^ sequence! !

!TCollateAccess methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:55'!
collateSequenceFor: aCharacter
	^self subclassResponsibility! !

!TCollateAccess methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:55'!
nextCollateSequenceFrom: anNKDStream
	| sequence nextSequence |
	sequence := self collateSequenceFor: anNKDStream next.
	[nextSequence := sequence first combinedWith: anNKDStream peek codePoint]
		whileTrue: [ anNKDStream next.
				sequence := nextSequence ].
	^sequence
	! !

Object subclass: #CollateElement
	instanceVariableNames: 'level1 additionalLevels sequences'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!CollateElement commentStamp: 'HenrikSperreJohansen 12/14/2015 11:13' prior: 0!
A CollateElement specifies the collate levels of a code point in a NFD normalized string, as well as the following code points for which a single collate element of a different value is the correct choice. !

!CollateElement methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 11:59'!
addFollowing: aCodePoint
	sequences ifNil: [sequences := SmallDictionary new. ].
	sequences at: aCodePoint ifAbsentPut: [ self class new ]! !

!CollateElement methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 12:00'!
combinedWith: aCodePoint
	^sequences ifNotNil: [ sequences at: aCodePoint ifAbsent: nil]! !

!CollateElement methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 12:09'!
isVariableWeighting
	^level1 negative! !

!CollateElement methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 12:08'!
level: anInt
	anInt < 1 ifTrue: [ self error: 'Collate level must be positive' ].
	anInt = 1 ifTrue: [ ^level1 abs].
	^(additionalLevels >> (anInt - 2)) bitAnd: 16rFF! !

!CollateElement methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 11:54'!
levels: anArray
	level1 := anArray first.
	"Other levels are standardized to never be more than 8 bits"
	additionalLevels := 0.
	2 to: anArray size do: [:ix | additionalLevels := additionalLevels + (anArray at: ix) << (ix - 2 * 8) ]! !

Object subclass: #CollateStrategy
	instanceVariableNames: 'cetData'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!CollateStrategy commentStamp: 'HenrikSperreJohansen 12/14/2015 13:19' prior: 0!
A SortStrategy defines  user-customizable parameters for how strings should be sorted.!

!CollateStrategy class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:31'!
collateTable: aCollateTable
	"Use a strategy for collating that respects the given collate table"
	^self basicNew collateTable: aCollateTable! !

!CollateStrategy class methodsFor: 'defaults' stamp: 'HenrikSperreJohansen 12/14/2015 15:30'!
default
	^Default ifNil: [ Default :=  self collateTable: NeoUnicodeDUCETData default ]! !

!CollateStrategy class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:35'!
forLocale: aLocale
	^self collateTable: (NeoUnicodeDUCETData forLocale: aLocale)! !

!CollateStrategy class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:31'!
new
	^self collateTable: NeoUnicodeDUCETData default! !

!CollateStrategy methodsFor: 'collating' stamp: 'HenrikSperreJohansen 12/14/2015 13:15'!
collate: aCollateStream to: anotherCollateStream at: aCollateLevel
	"A very basic collate strategy"
	^(aCollateStream peek level: aCollateLevel) <= (anotherCollateStream peek level: aCollateLevel)! !

!CollateStrategy methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:29'!
collateTable
	^cetData! !

!CollateStrategy methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 15:32'!
collateTable: aCollateTable
	cetData := aCollateTable! !

!CollateStrategy methodsFor: 'private  - sort delegation' stamp: 'HenrikSperreJohansen 12/14/2015 13:19'!
collectRanges: aCollateStreamCollection oldRanges: startRanges at: aLevel
	"Put this here, rather than in sorter, as I am uncertain whether collate strategies exist where ranges will have to be defined differently. 
	For now; the ranges are the elements after a sort that have the same collate value at the current level"
	|newRanges|
	newRanges := startRanges class new: startRanges size.
	startRanges do: [ :range | |rangeStart lastVal|
			rangeStart := range first.
			lastVal := (aCollateStreamCollection at: range first) peek level: aLevel.
			range first + 1 to: range last do: [:i | lastVal ~= (lastVal := (aCollateStreamCollection at: i) peek level: aLevel) ifTrue: [rangeStart - i > 0 ifTrue: [newRanges add: (rangeStart to: i)].
					rangeStart := i ]]].
		^newRanges! !

Object subclass: #NeoUnicodeDUCETData
	uses: TCollateAccess
	instanceVariableNames: 'collateTable'
	classVariableNames: 'DUCET EmptyElement'
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!NeoUnicodeDUCETData commentStamp: 'HenrikSperreJohansen 12/14/2015 11:01' prior: 0!
This class holds the collate data in the DUCET table (http://www.unicode.org/Public/UCA/latest/allkeys.txt). !

NeoUnicodeDUCETData class
	uses: TCollateAccess classTrait
	instanceVariableNames: ''!

!NeoUnicodeDUCETData class methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 12:27'!
addCEFrom: allKeysStream to: collateTable

| codePointHexStringCollection codePointCollection element levels |
codePointHexStringCollection := (allKeysStream upTo: $;) splitOn: Character space.
codePointCollection := codePointHexStringCollection collect: [ :each | Number readFrom: each base: 16 ].
"Find or create base"
	element := collateTable at: codePointCollection first ifAbsentPut: [ CollateElement new].
	codePointCollection allButFirstDo: [:codePoint | element := element addFollowing: codePoint ].
	allKeysStream upTo: $[.
	levels := ((allKeysStream upTo: $]) splitOn: $.) collect: [ :each | Number readFrom: each base: 16 ]..
	element levels: levels.
	^element! !

!NeoUnicodeDUCETData class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 11:24'!
default
	^DUCET ifNil: [self importDUCET ]! !

!NeoUnicodeDUCETData class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:36'!
forLocale: aLocale
	aLocale ifNil: [^self default].
	^NeoUnicodeLocaleCETData forLocale: aLocale! !

!NeoUnicodeDUCETData class methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 11:31'!
importDUCET
	|allKeysStream collateTable|
	allKeysStream := ZnClient new streaming:true; get: 'http://www.unicode.org/Public/UCA/latest/allkeys.txt'.
	collateTable := Dictionary new.
	[allKeysStream atEnd] whileFalse: [ self addCEFrom: allKeysStream to: collateTable].
	
	^DUCET := self new initializeFrom: collateTable! !

!NeoUnicodeDUCETData class methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 15:26'!
initialize
	EmptyElement := CollateElement new levels: #(0 0 0 0)! !

!NeoUnicodeDUCETData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:49'!
collateSequenceFor: aCharacter
	aCharacter ifNil: [ ^EmptyElement].
	^collateTable at: aCharacter codePoint ifAbsent: EmptyElement! !

Object subclass: #NeoUnicodeLocaleCETData
	uses: TCollateAccess
	instanceVariableNames: 'collateTable'
	classVariableNames: 'LocaleCETs'
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!NeoUnicodeLocaleCETData commentStamp: 'HenrikSperreJohansen 12/14/2015 11:02' prior: 0!
Instances of this class is intended to hold Locale-specific alterations to the standard DUCET table, as can be read from the CLDR base.!

NeoUnicodeLocaleCETData class
	uses: TCollateAccess classTrait
	instanceVariableNames: ''!

!NeoUnicodeLocaleCETData class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
createNewFor: aLocale
	^self new initializeFrom: (self loadCETDataFor: aLocale)! !

!NeoUnicodeLocaleCETData class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
forLocale: aLocale
	^LocaleCETs at: aLocale ifAbsentPut: [ self createNewFor: aLocale ]! !

!NeoUnicodeLocaleCETData class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
loadCETDataFor: aLocale
	"Find and parse the CLDR collate alterations for aLocale"
	^self notYetImplemented! !

!NeoUnicodeLocaleCETData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:53'!
collateSequenceFor: aCharacter
	aCharacter
		ifNil: [ ^ NeoUnicodeDUCETData default collateSequenceFor: aCharacter ].
	^ collateTable
		at: aCharacter codePoint
		ifAbsent: [ NeoUnicodeDUCETData default collateSequenceFor: aCharacter ]! !

Object subclass: #NeoUnicodeSorter
	instanceVariableNames: 'strategy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!NeoUnicodeSorter commentStamp: 'HenrikSperreJohansen 12/14/2015 12:29' prior: 0!
NeoUnicodeSorter sorts strings according to a defined collate table and strategy.!

!NeoUnicodeSorter class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:34'!
forLocale: aLocale
	^self strategy: (CollateStrategy forLocale: aLocale)! !

!NeoUnicodeSorter class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 12:56'!
new
	^self strategy: CollateStrategy default! !

!NeoUnicodeSorter class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:27'!
strategy: aCollateStrategy
	^self basicNew initializeFor: aCollateStrategy! !

!NeoUnicodeSorter methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 15:28'!
initialize: aStrategy
	strategy := aStrategy! !

!NeoUnicodeSorter methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 12/14/2015 15:28'!
sort: aStringCollection
	| collateStreams |
	collateStreams := aStringCollection collect: [ :aString | UnicodeCollateStream on: aString collateTable: strategy collateTable ].
	self sort: collateStreams from: 1 to: collateStreams size.
	^collateStreams collect: #contents! !

!NeoUnicodeSorter methodsFor: 'private - sort' stamp: 'HenrikSperreJohansen 12/14/2015 18:56'!
sort: collateStreams from: start to: end
	| currentLevel remainingRanges |
	currentLevel := 1.
	remainingRanges := Set with: (start to: end).
	collateStreams
		mergeSortFrom: start
		to: end
		by: [ :a :b | strategy collate: a to: b at: currentLevel ].
	[ remainingRanges
		=
			(remainingRanges := strategy
				collectRanges: collateStreams
				oldRanges: remainingRanges
				at: currentLevel)
		ifTrue:
			[ "No more differences on this level, advance to next" 			remainingRanges
				do:
					[ :range | 
					range do: [ :i | (collateStreams at: i) reset ].
					currentLevel := currentLevel + 1 ].
			remainingRanges isEmpty ]
		whileFalse:
			[ remainingRanges
				do: [ :range | range do: [ :i | (collateStreams at: i) next ] ].
			remainingRanges
				do:
					[ :range | 
					collateStreams
						mergeSortFrom: range first
						to: range last
						by: [ :a :b | strategy collate: a to: b at: currentLevel ] ] ]	]! !

Object subclass: #UnicodeCollateStream
	instanceVariableNames: 'strategy nkdStream ceSeq seqIndex nextSequence'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Sorting'!

!UnicodeCollateStream methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 14:58'!
contents
	^nkdStream contents! !
NeoUnicodeDUCETData initialize!
